// © 2001 Michael Pfeiffer

/*
#include <be/interface/Menu.h>
#include <be/interface/MenuItem.h>
#include <be/interface/PopUpMenu.h>
#include <be/interface/MenuField.h>
#include <be/interface/CheckBox.h>
#include <be/interface/StringView.h>
#include <be/support/String.h>
#include <be/app/MessageFilter.h>
*/
#include <Menu.h>
#include <MenuItem.h>
#include <PopUpMenu.h>
#include <MenuField.h>
#include <CheckBox.h>
#include <StringView.h>
#include <String.h>
#include <MessageFilter.h>
#include <View.h>
#include <Button.h>
#include <Window.h>

#define BUILDING_ADDON
#include "BecassoAddOn.h"
#include "AddOnSupport.h"
#ifdef IMAGEMANIP_WEAK_LINKAGE
#undef _IMPEXP_IMAGEMANIP // for weak linkage
#endif
#include "ImageManip.h"
#include "BBitmapAccessor.h"

#define MSG_FILTER_SELECTED     'fsel'
#define MSG_FILTER_PREVIEW      'prev'
#define MSG_APPLY_TO_SELECTION  'aply'

#ifdef MANIPULATOR
#define WINDOW_TITLE       "Manipulate "
#define NAME               "Filter"
#define MENU_FIELD         "Filter:"
#define POPUP_NAME         "Select Filter"
#define WRAPPER_TYPE  	   BECASSO_FILTER
#define PREVIEW            PREVIEW_2x2
#define APPLY_TO_SELECTION "Apply filter to selection too"
#endif

#ifdef CONVERTER
#define WINDOW_TITLE       "Convert "
#define NAME               "Generator"
#define MENU_FIELD         "Generator:"
#define POPUP_NAME         "Select Generator"
#define WRAPPER_TYPE  	   BECASSO_TRANSFORMER
#define PREVIEW            PREVIEW_2x2
#define APPLY_TO_SELECTION "Apply transformer to selection too"
#endif

class WrapperView;

//static WrapperView	*window = NULL;
static int32             mAddonId = -1;
static bool              mApplyToSelection = false;

class WrapperView : public BView
{
	BView       *mView;
	BMenuField  *mMenuField;
	int          mTop, 
	             mHeight, 
	             mWidth; 
	BStringView *mStatus;
	BString      mName;
	BString      mFilter;

	BMessage* SelectionMessage (int32 addon_id, const char* addon_name);	
	void SetupFilters(BMenu* sub_menu);
	void SetupView(const char* name);	
public:
	WrapperView (BRect rect);
	virtual	~WrapperView ();

	void AttachedToWindow();
	void DetachedFromWindow();
	void MessageReceived(BMessage* msg);
	void RemoveView();
	void SetStatus(const char* text);
	void SetupConfigView(int32 addon, const char* name);
	void SetName(const char* name)   { mName = name; }
	void SetFilter(const char* name) { mFilter = name; }
	void UpdateTitle();
};

void WrapperView::AttachedToWindow() {
	fprintf(stderr, "WrapperView::AttachedToWindow %x " WINDOW_TITLE "\n", this);
	BMenu* m = mMenuField->Menu();
	const int n = m->CountItems();
	for (int i = 0; i < n; i++) {
		BMenu* n = m->SubmenuAt(i);
		if (n) n->SetTargetForItems(this);
	}
}

void WrapperView::DetachedFromWindow() {
	fprintf(stderr, "WrapperView::DetachedFromWindow %x " WINDOW_TITLE "\n", this);
}

void WrapperView::SetupView(const char* name) {
	BView* v = FindView(name);
	if (v) v->SetResizingMode(B_FOLLOW_LEFT | B_FOLLOW_RIGHT | B_FOLLOW_BOTTOM);
}

WrapperView::WrapperView (BRect rect) 
	: BView (rect, "wrapper_view", B_FOLLOW_LEFT |B_FOLLOW_TOP, B_WILL_DRAW) {
	fprintf(stderr, "WrapperView: " WINDOW_TITLE "\n");
	mView = NULL;
	BMenu *sub_menu = new BPopUpMenu(POPUP_NAME);
	mHeight = rect.IntegerHeight(); mWidth = rect.IntegerWidth();
	mHeight = 0;
	mWidth = 270;
	ResizeTo(mWidth, mHeight + 16);
//	SetupView("AddonStatusBar"); SetupView("AddonStop"); SetupView("AddonInfo");
//	SetupView("AddonApply");	

//	mStatus = new BStringView(BRect(5, mHeight, 265, mHeight + 16), "", "Status", B_FOLLOW_LEFT | B_FOLLOW_BOTTOM);
//	AddChild(mStatus);
//	mHeight += 20;


	SetupFilters(sub_menu);
	mTop = 5;
	mMenuField = new BMenuField(BRect(5, mTop, 200, mTop+30), NAME, MENU_FIELD, sub_menu);
	mMenuField->SetDivider(0);
	sub_menu->SetRadioMode(false);
	AddChild(mMenuField);
	AddChild(new BButton(BRect(205, mTop, 265, mTop + 20), "", "Preview", new BMessage(MSG_FILTER_PREVIEW)));
	mTop += 32; 
	AddChild(new BCheckBox(BRect(5, mTop, 265, mTop + 20), "", APPLY_TO_SELECTION, new BMessage(MSG_APPLY_TO_SELECTION))); 
	mTop += 22;
	ResizeTo(mWidth, mHeight + mTop);
};


WrapperView::~WrapperView() {
	fprintf(stderr, "~WrapperView: " WINDOW_TITLE "\n");
//	window = NULL;
}


BMessage* WrapperView::SelectionMessage(int32 addon_id, const char *addon_name) {
	BMessage* msg = new BMessage(MSG_FILTER_SELECTED);
	msg->AddInt32("addon_id", addon_id);
	msg->AddString("addon_name", addon_name);
	return msg;
}

#if 1
void WrapperView::SetupFilters(BMenu* filtermenu) {
	// This code is stolen from BePhotoMagic!
	BMenu *tempmenu;
	BMenuItem *tempitem;
 	uint16 categorycount;
	BMessage *msg;
	
	if(filtermenu->CountItems()!=0)
		filtermenu->RemoveItems(0,filtermenu->CountItems());

	// Create bitmap accessor object to pass along, so that we
	// only get the add-ons that support this bitmap in the menu
	BBitmap *bmp = new BBitmap(BRect(0,0,320,240),B_RGBA32);
	BBitmapAccessor* accessor = Image_CreateBBitmapAccessor(bmp);
	accessor->SetDispose(false);
	
	const char *addonName;
	const char *addonInfo;
	const char *addonCategory;
	int32 addonVersion;
	image_addon_id *outList;
	int32 outCount = 0;
	
	// Get Manipulator filters
#ifdef MANIPULATOR
	if (Image_GetManipulators(accessor, NULL, &outList, &outCount) == B_OK)
#endif
#ifdef CONVERTER
	if (Image_GetConverters(accessor, NULL, &outList, &outCount) == B_OK)
#endif
	{
//		fprintf(stderr, "ImageManipWrapper\n");
		for (int i = 0; i < outCount; ++i)
		{
			// Get plugin
			if (Image_GetAddonInfo(outList[i], &addonName, &addonInfo,
					&addonCategory, &addonVersion) == B_OK)
			{
//				fprintf(stderr, "%s\n", addonName);
				// Does it belong to a new category
				categorycount=filtermenu->CountItems();

				if(categorycount!=0)
				{	
					// New category?
					tempitem=filtermenu->FindItem(addonCategory);
					if(tempitem!=NULL)
					{
						tempmenu=tempitem->Submenu();
						
						// nothing new, so just add the item
						msg = SelectionMessage(outList[i], addonName);
	
						// Add at position 0 so it's sorted properly
						tempmenu->AddItem(new BMenuItem(addonName, msg),0);
					}
					else
					{	// New category - add category and item
						msg = SelectionMessage(outList[i], addonName);
	
						tempmenu=new BMenu(addonCategory);
	
						// Add at position 0 so it's sorted properly
						tempmenu->AddItem(new BMenuItem(addonName, msg),0);
						filtermenu->AddItem(tempmenu);
					}
				}
				else
				{	
					// Nothing's been added yet, so add the category and
					// the filter into it						
					msg = SelectionMessage(outList[i], addonName);

					tempmenu=new BMenu(addonCategory);

					// Add at position 0 so it's sorted properly
					tempmenu->AddItem(new BMenuItem(addonName, msg),0);
					filtermenu->AddItem(tempmenu);
				}
			}
//			if (i == 0) PostMessage(msg);
			// if (i == outCount-1) SetupConfigView(outList[i], addonName);
		}
		delete[] outList;
	}
	delete bmp;
	delete accessor;
}
#else
void WrapperView::SetupFilters(BMenu* sub_menu) {
	// Create bitmap accessor object to pass along, so that we
	// only get the add-ons that support this bitmap in the menu
	BBitmap bitmap(BRect(0, 0, 100, 20), B_RGB32);
	BBitmapAccessor* accessor = Image_CreateBBitmapAccessor(&bitmap);
	if (accessor == NULL) {
		sub_menu->SetEnabled(false); return;
	}
	accessor->SetDispose(false);

	// Create sub menu with image manipulators
	const char *addonName;
	const char *addonInfo;
	const char *addonCategory;
	int32 addonVersion;
	image_addon_id *outList;
	int32 outCount = 0;
	BMenuItem *item;
#ifdef MANIPULATOR
	if (Image_GetManipulators(accessor, NULL, &outList, &outCount) == B_OK)
#endif
#ifdef CONVERTER
	if (Image_GetConverters(accessor, NULL, &outList, &outCount) == B_OK)
#endif
	{
		mAddonId = outList[0];
		
		for (int i = outCount-1; i >= 0; --i)
		{
			if (Image_GetAddonInfo(outList[i], &addonName, &addonInfo,
					&addonCategory, &addonVersion) == B_OK)
			{
				BMessage *msg = new BMessage(MSG_FILTER_SELECTED);
				msg->AddInt32("addon_id", outList[i]);
				msg->AddString("addon_name", addonName);
				item = new BMenuItem(addonName, msg);
				sub_menu->AddItem(item);
				if (i == outCount-1) PostMessage(msg);
			}
		}
		delete[] outList;
	}
	if (outCount == 0)
		sub_menu->SetEnabled(false);
	else 
	{
		sub_menu->ItemAt((int32)0)->SetMarked(true);
	}
	delete accessor;
}
#endif

void WrapperView::RemoveView() {
	if (mView) {
		RemoveChild(mView);
		delete mView; mView = NULL;
	}
}


void WrapperView::SetupConfigView(int32 addon, const char* name) {
	mAddonId = addon;
	RemoveView();
	int width;
	BMessage ioExt, cap;
	status_t s = Image_GetConfigurationMessage(mAddonId, &ioExt, &cap);
	if (Image_MakeConfigurationView(mAddonId, &ioExt, &mView) == 0 && mView) {
		AddChild(mView); 
		int width = max_c(mWidth, mView->Bounds().Width()) + 10;
		mView->MoveTo((width-mView->Bounds().Width()) / 2, mTop);
		ResizeTo(width, mHeight + mTop + mView->Bounds().Height());
	} else {
		ResizeTo(mWidth, mTop + mHeight);
		width = mWidth;
	}
	SetFilter(name);
	UpdateTitle();
}


void WrapperView::UpdateTitle() {
	BString title;
	if (mFilter == "") {
		title = WINDOW_TITLE;
	} else {
		title = mFilter;
		title += " ";
	}
	title += mName;
	// SetTitle(title.String());
}

void WrapperView::MessageReceived(BMessage* msg) {
//	fprintf(stderr, "WrapperView::MessageReceived %4.4s\n", &msg->what);
	switch(msg->what) {
		case MSG_FILTER_SELECTED: {
			BString name;
			SetStatus("Status");
			msg->FindString("addon_name", &name);
			SetupConfigView(msg->FindInt32("addon_id"), name.String());
			Window()->PostMessage(ADDON_RESIZED);
			break; }
		case MSG_FILTER_PREVIEW:
//			aPreview();
			addon_preview();
			break;
		case MSG_APPLY_TO_SELECTION: {
			int32 s;
				if (B_OK == msg->FindInt32("be:value", &s)) {
					mApplyToSelection = (bool)s;
				}
			}
			break;
		default:
			BView::MessageReceived(msg);
	}
}


void WrapperView::SetStatus(const char* text) {
//	mStatus->SetText(text);
}


status_t addon_init (uint32 index, becasso_addon_info *info)
{
	fprintf(stderr, "addon_init " WINDOW_TITLE "\n");
	strcpy (info->name,        "LibImageManip Wrapper");
	strcpy (info->author,      "Michael Pfeiffer");
	strcpy (info->copyright,   "© 2001 under GPL");
	strcpy (info->description, "Wrapper to LibImageManip");
	info->type              = WRAPPER_TYPE;
	info->index				= index;
	info->version			= 0;
	info->release			= 6;
	info->becasso_version	= 2;
	info->becasso_release	= 0;
	info->does_preview		= PREVIEW;
	info->flags             = 0;
#if defined(MANIPULATOR) || defined(IMAGEMANIP_WEAK_LINKAGE)
	// Note: We assume that the filter is loaded first by Becasso,
	//       then the transformer is loaded and both are loaded
	//       or none.
	//       Imagelibmanip is initialized once here.
	//       When we exit we assume that the transformer is called
	//       last and we close Imagelibmanip there.
	//       We must also assume that there is no other filter that
	//       uses Imagelibmanip otherwise it is likely that Becasso
	//       will crash when the add-ons are exited (addon_exit).
	//
	// When IMAGEMANIP_WEAK_LINKAGE is defined we solve this problem
	// as Imagelibmanip is loaded in each wrapper add-on.       

	// Check the library's version
	int32 curVersion;
	int32 minVersion;

	// Initialize the image manipulation library
	if (Image_Init(NULL) != B_OK)
	{
		puts("Cannot initialize libimagemanip.so");
		return B_ERROR;
	}

	// Check the library's version
	puts(Image_Version(&curVersion, &minVersion));
	if (curVersion < IMAGE_LIB_MIN_VERSION)
	{
		puts("Need a newer version of libimagemanip.so");
		return B_ERROR;
	}
	if (minVersion > IMAGE_LIB_CUR_VERSION)
	{
		puts("Need an older version of libimagemanip.so");
		return B_ERROR;
	}
#endif
	return (0);
}

status_t addon_exit (void)
{
	fprintf(stderr, "addon_exit " WINDOW_TITLE "\n");
	// We have to remove the config view from Imagelibmanip otherwise we
	// crash when the destructor of this window is called!
//	window->Lock();
//	window->RemoveView();
//	window->Unlock();
	// Shutdown the image manipulation library
#if defined(MANIPULATOR) || defined(IMAGEMANIP_WEAK_LINKAGE)
	// Note: We assume that the addon_exit of the transformer is called
	//       after the filter add-on.
	Image_Shutdown();
#endif	
	return (0);
}

status_t addon_make_config (BView **view, BRect rect)
{
	fprintf(stderr, "addon_config " WINDOW_TITLE "\n");
/*
	window->Lock();
	window->SetName (name);
	window->UpdateTitle ();
	if (window->IsHidden())
		window->aShow (client);
	else
		window->aActivate (client);
	window->SetStatus("Status");
	window->Unlock();
	return (0);
*/
	if (*view == NULL) {
		*view = new WrapperView(rect);
	}
	return B_OK;
}

status_t addon_close ()
{
	fprintf(stderr, "addon_close " WINDOW_TITLE "\n");
//	window->aClose();
	return (0);
}

static void SetStatus(const char* text) {
//	window->Lock();
//	window->SetStatus(text);
//	window->Unlock();
}


static void CopySelection(BBitmap* src, BBitmap* dst, int x, int y) {
	dst->Lock();
	BView* view = new BView(dst->Bounds(), "tmp", 0, 0);
	dst->AddChild(view);
	view->DrawBitmap(src, BPoint(-x, -y));
	view->Sync();
	dst->RemoveChild(view);
	delete view;
	dst->Unlock();
}


status_t process (Layer *inLayer, Selection *inSelection, 
			 Layer **outLayer, Selection **outSelection, int32 mode,
			 BRect * frame, bool final, BPoint /* point */, uint32 /* buttons */)
{
	int error = ADDON_OK;
	if (mAddonId == -1) return ADDON_UNKNOWN;
	BRect bounds = inLayer->Bounds();

//	printf ("Bounds: ");
//	bounds.PrintToStream();
//	printf ("Frame:  ");
//	frame->PrintToStream();
//	fprintf(stderr, "mode = %d final = %d\n", mode, final);
//	fprintf(stderr, "in %x out %x in %x out %x\n", inLayer, *outLayer, inSelection, *outSelection);

// #ifdef MANIPULATOR
	if (*outLayer == NULL && mode == M_DRAW) {
		*outLayer = new Layer (*inLayer);
	}
	if (*outSelection == NULL && mode == M_DRAW) {
		*outSelection = inSelection;
	}
	
	if (mode == M_SELECT)
	{
		if (inSelection)
		{
		 	if (*outSelection == NULL)
				*outSelection = new Selection (*inSelection);
		}
		else	// No Selection to filter!
			return (0);
	}
	if (*outLayer)
		(*outLayer)->Lock();
		
	if (*outSelection)
		(*outSelection)->Lock();
// #endif

//	fprintf(stderr, "in %x out %x in %x out %x\n", inLayer, *outLayer, inSelection, *outSelection);
		
	uint32 h = bounds.IntegerHeight() + 1;
	uint32 w = bounds.IntegerWidth() + 1;

	if (final)
	{
		addon_start();
	} 
	else
	{
		return error;
	} 
	
	BMessage ioExt, m;
	Image_GetConfigurationMessage(mAddonId, &ioExt, &m);
//	ioExt.PrintToStream();
//	m.PrintToStream();
	if (frame->IntegerWidth() == 0 || frame->IntegerHeight() == 0) {
		ioExt.AddRect("selection_rect", bounds);
	} else {
		// frame->PrintToStream();
		ioExt.AddRect("selection_rect", *frame);
	}
	printf("ioExt:\n");
	ioExt.PrintToStream();
	printf("ioCapability:\n");
	m.PrintToStream();
	fflush(stdout);
	switch (mode)
	{
	#ifdef MANIPULATOR // Becasso Filter
		case M_DRAW:
		{
			// Create bitmap accessor object
			memcpy((*outLayer)->Bits(), inLayer->Bits(), inLayer->BitsLength());
			Layer* layer;
			if (inSelection) {
				layer = new Layer(*inLayer);
				memcpy(layer->Bits(), inLayer->Bits(), inLayer->BitsLength());
			} else {
				layer = *outLayer;
			}
			BBitmapAccessor* src_accessor = Image_CreateBBitmapAccessor(layer);
			if (src_accessor == NULL) {
				error = ADDON_ABORT;;
				if (inSelection) delete layer;
				break;
			}
			src_accessor->SetDispose(false);
			
			// manipulate layer
			if (Image_Manipulate(mAddonId, src_accessor, &ioExt, false) != B_OK) {
				SetStatus("Error: Could not manipulate layer!");
				error = ADDON_UNKNOWN;
			}
			delete src_accessor;
			
			// manipulate selection			
			if (error == ADDON_OK && mApplyToSelection && inSelection) {
				memcpy((*outSelection)->Bits(), inSelection->Bits(), inSelection->BitsLength());
				src_accessor = Image_CreateBBitmapAccessor(*outSelection);
				src_accessor->SetDispose(false);
				if (Image_Manipulate(mAddonId, src_accessor, &ioExt, false) != B_OK) {
					SetStatus("Error: Could not manipulate selection!");
				}
				delete src_accessor;
			}

			// return selection of layer only
			if (error == ADDON_OK && inSelection && *outSelection) {
				CutOrCopy(layer, layer, *outSelection, 0, 0, false);
				AddWithAlpha(layer, *outLayer, 0, 0);
			}

			if (inSelection) {
				delete layer;
			}			
		break;
		}
		case M_SELECT:
		{
			// Create bitmap accessor object
			memcpy((*outSelection)->Bits(), inSelection->Bits(), inSelection->BitsLength());
			BBitmapAccessor* src_accessor = Image_CreateBBitmapAccessor(*outSelection);
			if (src_accessor == NULL) {
				error = ADDON_ABORT;;
				break;
			}
			src_accessor->SetDispose(false);

			if (Image_Manipulate(mAddonId, src_accessor, &ioExt, false) < 0) {
//				window->Lock();
//				window->SetStatus("Error: Could not manipulate selection!");
//				window->Unlock();
			}
			delete src_accessor;
		break;
		}
	#endif
	
	
	#ifdef CONVERTER // Becasso Transformer
		case M_DRAW:
		{
			// Create bitmap accessor object
			Layer* layer;
			if (inSelection) {
				layer = new Layer(*inLayer);
				memcpy((*outLayer)->Bits(), inLayer->Bits(), inLayer->BitsLength());
			} else {
				layer = *outLayer;
			}


			BBitmapAccessor* src_accessor = Image_CreateBBitmapAccessor(inLayer);
			if (src_accessor == NULL) {
				error = ADDON_ABORT;;
				break;
			}
			src_accessor->SetDispose(false);

			BBitmapAccessor* dst_accessor = Image_CreateBBitmapAccessor();
			dst_accessor->SetDispose(true);
			
			// convert layer
			if (Image_Convert(mAddonId, src_accessor, dst_accessor, &ioExt, false) != B_OK) {
				SetStatus("Error: Could not convert layer!");
				error = ADDON_UNKNOWN;
			}
			delete src_accessor;
			
			int dx = -(bounds.IntegerWidth() - dst_accessor->Bounds().IntegerWidth()) / 2; 
			int dy = -(bounds.IntegerHeight() - dst_accessor->Bounds().IntegerHeight()) / 2; 
			CutOrCopy(dst_accessor->Bitmap(), layer, NULL, dx, dy, false);
			delete dst_accessor;
						
			// manipulate selection			
			if (error == ADDON_OK && mApplyToSelection && inSelection) {
				src_accessor = Image_CreateBBitmapAccessor(inSelection);
				src_accessor->SetDispose(false);
				dst_accessor = Image_CreateBBitmapAccessor();
				dst_accessor->SetDispose(true);
				if (Image_Convert(mAddonId, src_accessor, dst_accessor, &ioExt, false) != B_OK) {
					SetStatus("Error: Could not manipulate selection!");
				}
				delete src_accessor;
				CopySelection(dst_accessor->Bitmap(), *outSelection,dx, dy);
				delete dst_accessor;
			}

			// return selection of layer only
			if (error == ADDON_OK && inSelection && *outSelection) {
				CutOrCopy(layer, layer, *outSelection, 0, 0, false);
				AddWithAlpha(layer, *outLayer, 0, 0);
			}

			if (inSelection) {
				delete layer;
			}

		break;
		}
		case M_SELECT:
		{
			// Create bitmap accessor object
			BBitmapAccessor* src_accessor = Image_CreateBBitmapAccessor(inSelection);
			if (src_accessor == NULL) {
				error = ADDON_ABORT;;
				break;
			}
			src_accessor->SetDispose(false);
			BBitmapAccessor* dst_accessor = Image_CreateBBitmapAccessor();
			dst_accessor->SetDispose(true);
			
			if (Image_Convert(mAddonId, src_accessor, dst_accessor, &ioExt, false) != B_OK) {
				SetStatus("Error: Could not convert selection!");
				error = ADDON_ABORT;;
			} else {
				int dx = -(bounds.IntegerWidth() - dst_accessor->Bounds().IntegerWidth()) / 2; 
				int dy = -(bounds.IntegerHeight() - dst_accessor->Bounds().IntegerHeight()) / 2; 
				CopySelection(dst_accessor->Bitmap(), *outSelection, dx, dy);
			}
			delete src_accessor; delete dst_accessor;
		break;
		}
#endif 
	
	default:
		fprintf (stderr, "Wrapper: Unknown mode\n");
		error = ADDON_UNKNOWN;
	}

// #ifdef MANIPULATOR
	if (*outSelection)
		(*outSelection)->Unlock();
	if (*outLayer)
		(*outLayer)->Unlock();
// #endif
	
	if (final)
	{
		addon_done();
	}
	return (error);
}
